; Berechnung und Ausgabe extrem langer
; Fibonaccizahlen mit sehr vielen Dezimalstellen

.include "m328Pdef.inc"

.equ F_CPU = 16000000                            ; Systemtakt in Hz
.equ BAUD  = 76800                               ; Baudrate

; Berechnungen
.equ UBRR_VAL   = ((F_CPU+BAUD*8)/(BAUD*16)-1)  ; clever runden
.equ BAUD_REAL  = (F_CPU/(16*(UBRR_VAL+1)))     ; Reale Baudrate
.equ BAUD_ERROR = ((BAUD_REAL*1000)/BAUD-1000)  ; Fehler in Promille

.if ((BAUD_ERROR>10) || (BAUD_ERROR<-10))       ; max. +/-10 Promille Fehler
  .error "Systematischer Fehler der Baudrate grsser 1 Prozent und damit zu hoch!"
.endif

; Druckausgabe ein/aus
#define PRINT      

.macro AVRTEST_PUTCHAR  
loop:
    lds     tmp2, UCSR0A
    sbrs    tmp2, UDRE0
    rjmp loop
    sts     UDR0, arg0    
.endm

; maximale Anzahl Dezimalstellen der Fibonaccizahlen (1..255)
.equ maxDigits = 80

; Registerbelegung
.def tmp2      = r15
.def tmp       = r16
.def count     = r17
.def digit0    = r18
.def digit1    = r19
.def nDigits   = r20
.def minus10   = r21
.def pageXor   = r22
.def arg0      = r24
.def arg1      = r25
; X, Y, und Z werden benutzt

.dseg
; pageA und pageB mssen auf ganzen 0x100er Adressen liegen
; damit die vereinfachte Adressierung funktioniert

.org 0x100
pageA:  .byte maxDigits

.org 0x200
pageB:  .byte maxDigits

.cseg
.org 0
reset:
    ldi tmp, high(ramend)
    out sph, tmp
    ldi tmp, low(ramend)
    out spl, tmp

start:

#ifdef PRINT
    // UART einstellen
    ldi     tmp, HIGH(UBRR_VAL)
    sts     UBRR0H, tmp
    ldi     tmp, LOW(UBRR_VAL)
    sts     UBRR0L, tmp

    ; Frameformat: 8 Bit, no parity
    ldi     tmp, 0x06
    sts     UCSR0C, tmp
    ldi     tmp, (1<<TXEN0)
    sts     UCSR0B, tmp

    ldi arg0, '\r'
    AVRTEST_PUTCHAR
    ldi arg0, '\n'
    AVRTEST_PUTCHAR
#endif

    ; pageA mit '0' fllen, es wird direkt mit ASCII gerechnet
    ; A = 0
    clr XL
    ldi XH, high(pageA)
    ldi digit0, '0'         
    ldi count, maxDigits
loopA:                      
    st X+, digit0
#ifdef PRINT
    mov arg0, digit0
    AVRTEST_PUTCHAR
#endif
    dec count
    brne loopA

#ifdef PRINT
    ; Ausgabe fib(0)
    ldi arg0, '\r'
    AVRTEST_PUTCHAR
    ldi arg0, '\n'
    AVRTEST_PUTCHAR
#endif

    ; pageB mit '0' fllen, es wird direkt mit ASCII gerechnet
    ; B=1
    clr XL
    ldi XH, high(pageB)
    ldi digit0, '0'
    ldi count, maxDigits-1
loopB:
    st X+, digit0
    dec count
    brne loopB

    ldi digit0, '1'
    st X, digit0

    ; weitere Initialisierungen
    ldi minus10, -('9'+1)
    ldi pageXor, high(pageA) ^ high(pageB)

    ; zu berechnende Stellenzahl festlegen, anfangs reicht 1, mit grer
    ; werdenden Zahlen wird ndigits entsprechend angepasst
    ldi nDigits, 1

    ; X zeigt auf A, Y auf B
    ldi XH, high(pageA)
    ldi YH, high(pageB)

    ; Schleife ber alle Fibonacci-Zahlen
loop_Fib:
    ldi XL, maxDigits
    ldi YL, maxDigits
    clc
    mov count, nDigits

    ; Schleife ber alle Ziffern
loop_Zif:
    ld digit0, -X
    ld digit1, -Y

    ; ASCII-Ziffern addieren
    adc digit0, digit1
    subi digit0, '0'

    ; probeweise 10 subtrahieren, um ggf. bertrag festzustellen
    mov tmp, digit0
    add tmp, minus10

    brcc kein_uebertrag1
    ; bertrag behandeln
    mov digit0, tmp
    ori digit0, '0'
kein_uebertrag1:
    st X, digit0
    dec count
    brne loop_Zif

#ifdef PRINT
    ; Ausgabe fib(i)
    in tmp, SREG        ; C-Flag speichern
    clr ZL
    mov ZH, YH
    ldi count, maxDigits
loop_putchar:
    ld arg0, Z+
    AVRTEST_PUTCHAR
    dec count
    brne loop_putchar
    ldi arg0, '\r'
    AVRTEST_PUTCHAR
    ldi arg0, '\n'
    AVRTEST_PUTCHAR
    out SREG, tmp       ; C-Flag wieder herstellen
#endif

    ; falls nach der Berechnung der Zahl ein bertrag verbleibt
    brcc kein_uebertrag2
    ; beim Erreichen der maximalen Stellenzahl Programm beenden
    cpi nDigits, maxDigits
    breq exit

    ; sonst Stellenzahl vergern
    inc nDigits
    ; und den bertrag in die nchsthhere Ziffer schreiben
    ldi digit0, '1'
    st -X, digit0
kein_uebertrag2:
    ; Zeiger auf A und B vertauschen
    eor XH, pageXor
    eor YH, pageXor    
    rjmp loop_Fib

exit:
    rjmp exit

